@@ -178,6 +178,7 @@ XrdNetPMark::Handle *XrdNetPMarkCfg::Begin(XrdSecEntity &client,
                                            const char   *app)
 {
    EPName("PMBegin");
+   XrdOucString altApp;
    int eCode, aCode;
 
 // If we need to screen out domains, do that
@@ -206,6 +207,23 @@ XrdNetPMark::Handle *XrdNetPMarkCfg::Begin(XrdSecEntity &client,
        return 0;
       }
 
+// Obtain the appname overridefrom the cgi
+//
+   if (cgi)                        // 01234567890123
+      {const char *apP = strstr(cgi, "pmark.appname=");
+       if (apP)
+          {apP += 14;
+           const char* aP = apP;
+           while(*aP && *aP != '&') aP++;
+           int apLen = aP - apP;
+           if (apLen > 0)
+              {altApp = "";
+               altApp.insert(apP, 0, apLen);
+               app = altApp.c_str();
+              }
+          }
+      }
+
 // Continue with successor function to complete the logic
 //
    XrdNetPMark::Handle handle(app, eCode, aCode);
@@ -983,6 +1001,14 @@ do{if (!strcmp("debug", val) || !strcmp("nodebug", val))
            return 1;
           }
 
+       if (!strcmp("fail", val) || !strcmp("nofail", val))
+          {noFail = (*val == 'n');
+           if (!(val = Config.GetWord()))
+              {eLog->Say("Config invalid: pmark defsfile locationnot specified");
+               return 1;
+              }
+          }
+
        if (*val == '/')
           {Cfg->defsFile = val;
            continue;
